from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger
from typing import List

import libcamera

from scicamera.lc_helpers import lc_unpack

_log = getLogger(__name__)


@dataclass
class CameraInfo:
    id: str
    """Unique identifier for this camera."""

    model: str
    """Model name for this camera."""

    location: str | None
    """Location of the camera, if known."""

    rotation: int | None
    """Rotation of the camera, if known."""

    size: tuple[int, int] | None
    """Size of the pixel array, if known."""

    @staticmethod
    def global_camera_info() -> List[CameraInfo]:
        """
        Return Id string and Model name for all attached cameras, one dict per camera,
        and ordered correctly by camera number. Also return the location and rotation
        of the camera when known, as these may help distinguish which is which.
        """
        return list(
            CameraInfo.from_lc_camera(cam)
            for cam in libcamera.CameraManager.singleton().cameras
        )

    @classmethod
    def from_lc_camera(cls, lc_camera: libcamera.Camera) -> CameraInfo:
        name_to_val = {"location": None, "rotation": None, "size": None, "model": None}
        _log.debug("Camera Property Keys: %s", lc_unpack(lc_camera.properties))
        for k, v in lc_camera.properties.items():
            renamed = k.name.lower()
            if renamed == "pixelarraysize":
                renamed = "size"
            if renamed in name_to_val:
                name_to_val[renamed] = v
            else:
                _log.debug("Unknown property %s: %s", k.name, v)
        name_to_val["id"] = lc_camera.id
        return cls(**name_to_val)

    @staticmethod
    def n_cameras() -> int:
        """Return the number of attached cameras."""
        return len(libcamera.CameraManager.singleton().cameras)

    def requires_camera(needed: int = 1) -> None:
        """Require a minimum number of cameras to be attached.

        Raises: RuntimeError if not enough cameras are found.
        """
        found = CameraInfo.n_cameras()
        if found < needed:
            msg = "{n} camera(s) required found {found} not found (need) (Do not forget to disable legacy camera with raspi-config)."
            _log.error(msg)
            raise RuntimeError(msg)



#!/usr/bin/python3
"""scicamera main classes"""
from __future__ import annotations

import logging
import selectors
import threading
from collections import deque
from dataclasses import replace
from functools import partial
from typing import Dict, List

import libcamera

import scicamera.formats as formats
from scicamera.actions import RequestMachinery
from scicamera.configuration import CameraConfig
from scicamera.controls import Controls
from scicamera.info import CameraInfo
from scicamera.lc_helpers import errno_handle, lc_unpack, lc_unpack_controls
from scicamera.preview import NullPreview
from scicamera.request import CompletedRequest, LoopTask
from scicamera.sensor_format import SensorFormat
from scicamera.tuning import TuningContext

_log = logging.getLogger(__name__)


class CameraManager:
    cameras: Dict[int, Camera]

    _instance: CameraManager = None

    @classmethod
    def singleton(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self.running = False
        self.cameras = {}
        self._lock = threading.Lock()
        self.cms = libcamera.CameraManager.singleton()

    def n_cameras_attached(self):
        return len(self.cms.cameras)

    def get_camera(self, idx: int):
        """Get the (lc) camera with the given index"""
        return self.cms.cameras[idx]

    def setup(self):
        self.thread = threading.Thread(target=self.listen, daemon=True)
        self.running = True
        self.thread.start()

    def close_all(self) -> int:
        n_closed = 0
        ids_to_close = list(self.cameras.keys())
        for idx in ids_to_close:
            # FIXME(meawoppl) - this calls back into self.cleanup()
            self.cameras[idx].close()
            n_closed += 1
        return n_closed

    def add(self, index: int, camera: Camera):
        with self._lock:
            self.cameras[index] = camera
            if not self.running:
                self.setup()

    def cleanup(self, index: int):
        flag = False
        with self._lock:
            del self.cameras[index]
            if self.cameras == {}:
                self.running = False
                flag = True
        if flag:
            self.thread.join()

    def listen(self):
        sel = selectors.DefaultSelector()
        sel.register(self.cms.event_fd, selectors.EVENT_READ, self.handle_request)

        while self.running:
            events = sel.select(0.05)
            for key, _ in events:
                callback = key.data
                callback()

        sel.unregister(self.cms.event_fd)

    def handle_request(self, flushid: int | None = None) -> int:
        """Handle requests"""
        n_flushed = 0
        with self._lock:
            for req in self.cms.get_ready_requests():
                if req.status != libcamera.Request.Status.Complete:
                    _log.warning("Unexpected request status: %s", req.status)
                    continue
                if req.cookie == flushid:
                    _log.warning("Flushing request.")
                    n_flushed += 1
                    continue

                camera_inst = self.cameras[req.cookie]
                cleanup_call = partial(
                    camera_inst.recycle_request, camera_inst.stop_count, req
                )
                self.cameras[req.cookie].add_completed_request(
                    CompletedRequest(
                        req,
                        replace(camera_inst.camera_config),
                        camera_inst.stream_map,
                        cleanup_call,
                    )
                )
        return n_flushed


class Camera(RequestMachinery):
    """Welcome to the Camera class."""

    def __init__(self, camera_num: int = 0, tuning=None):
        """Initialise camera system and open the camera for use.

        :param camera_num: Camera index, defaults to 0
        :type camera_num: int, optional
        :param tuning: Tuning filename, defaults to None
        :type tuning: str, optional
        :raises RuntimeError: Init didn't complete
        """
        super().__init__()
        self._cm = CameraManager.singleton()

        self._cm.add(camera_num, self)
        self.camera_idx = camera_num
        self._reset_flags()

        with TuningContext(tuning):
            self._open_camera()

    def _reset_flags(self) -> None:
        self.camera = None
        self.is_open = False
        self.camera_ctrl_info = {}
        self._preview = None
        self.camera_config = None
        self.streams = None
        self.stream_map = None
        self.started = False
        self.stop_count = 0
        self.camera_properties_ = {}
        self.controls = Controls(self)
        self.sensor_modes_ = None

    @property
    def info(self) -> CameraInfo:
        """Get camera info

        :return: Camera info
        :rtype: CameraInfo
        """
        return CameraInfo.from_lc_camera(self.camera)

    @property
    def camera_properties(self) -> dict:
        """Camera properties

        :return: Camera properties
        :rtype: dict
        """
        return {} if self.camera is None else self.camera_properties_

    @property
    def camera_controls(self) -> dict:
        return {
            k: (v[1].min, v[1].max, v[1].default)
            for k, v in self.camera_ctrl_info.items()
        }

    def __del__(self):
        """Without this libcamera will complain if we shut down without closing the camera."""
        if self.is_open:
            _log.warning(f"__del__ call responsible for cleanup of {self}")
            self.close()

    def _initialize_camera(self) -> None:
        """Initialize camera

        :raises RuntimeError: Failure to initialize camera
        """
        CameraInfo.requires_camera()

        self.camera = self._cm.get_camera(self.camera_idx)
        if self.camera is None:
            message = "Initialization failed."
            _log.error(message)
            raise RuntimeError(message)

        self.__identify_camera()
        self.camera_ctrl_info = lc_unpack_controls(self.camera.controls)
        self.camera_properties_ = lc_unpack(self.camera.properties)

        # The next two lines could be placed elsewhere?
        self.sensor_resolution = self.camera_properties_["PixelArraySize"]
        self.sensor_format = str(
            self.camera.generate_configuration([libcamera.StreamRole.Raw])
            .at(0)
            .pixel_format
        )

        _log.info("Initialization successful.")

    def __identify_camera(self):
        # TODO(meawoppl) make this a helper on the camera_manager
        for idx, address in enumerate(self._cm.cms.cameras):
            if address == self.camera:
                self.camera_idx = idx
                break

    def _open_camera(self) -> None:
        """Tries to open camera

        :raises RuntimeError: Failed to setup camera
        """
        self._initialize_camera()

        acq_code = self.camera.acquire()
        errno_handle(acq_code, "camera.acquire()")

        self.is_open = True
        _log.info("Camera now open.")

    @property
    def sensor_modes(self) -> list:
        """The available sensor modes

        When called for the first time this will reconfigure the camera
        in order to read the modes.
        """
        if self.sensor_modes_ is not None:
            return self.sensor_modes_

        raw_config = self.camera.generate_configuration([libcamera.StreamRole.Raw])
        raw_formats = raw_config.at(0).formats
        self.sensor_modes_ = []

        for pix in raw_formats.pixel_formats:
            name = str(pix)
            if not formats.is_raw(name):
                # Not a raw sensor so we can't deduce much about it. Quote the name and carry on.
                self.sensor_modes_.append({"format": name})
                continue
            fmt = SensorFormat(name)
            all_format = {}
            all_format["format"] = fmt
            all_format["unpacked"] = fmt.unpacked
            all_format["bit_depth"] = fmt.bit_depth
            for size in raw_formats.sizes(pix):
                cam_mode = all_format.copy()
                cam_mode["size"] = (size.width, size.height)
                temp_config = CameraConfig.for_preview(
                    camera=self, raw={"format": str(pix), "size": cam_mode["size"]}
                )
                self.configure(temp_config)
                frameDurationMin = self.camera_controls["FrameDurationLimits"][0]
                cam_mode["fps"] = round(1e6 / frameDurationMin, 2)
                cam_mode["crop_limits"] = self.camera_properties["ScalerCropMaximum"]
                cam_mode["exposure_limits"] = tuple(
                    [i for i in self.camera_controls["ExposureTime"] if i != 0]
                )
                self.sensor_modes_.append(cam_mode)
        return self.sensor_modes_

    # TODO(meawoppl) we don't really support previews, so change the language here
    def start_preview(self) -> None:
        """
        Start the preview loop.
        """
        if self._preview:
            raise RuntimeError("An event loop is already running")

        preview = NullPreview()
        preview.start(self)
        self._preview = preview

    def stop_preview(self) -> None:
        """Stop preview the preview thread"""
        if not self._preview:
            return

        self._preview.stop()
        self._preview = None

    def close(self) -> None:
        """Close camera

        :raises RuntimeError: Closing failed
        """
        self.stop_preview()
        if not self.is_open:
            return

        self.stop()
        code = self.camera.release()
        errno_handle(code, "camera.release()")

        self._cm.cleanup(self.camera_idx)
        self.is_open = False
        self.streams = None
        self.stream_map = None
        self.camera = None
        self.camera_ctrl_info = None
        self.camera_config = None
        self._preview_configuration = None
        self.allocator = None
        _log.info("Camera closed successfully.")

    def recycle_request(self, stop_count: int, request: libcamera.Request) -> None:
        """Recycle a request.

        :param request: request
        :type request: libcamera.Request
        """
        if not self.camera:
            _log.warning("Can't recycle request, camera not open")
            return

        if stop_count != self.stop_count:
            _log.warning("Can't recycle request, stop count mismatch")
            return

        request.reuse()

        # This is where controls on the camera get updated
        # TODO(meawoppl) - attach a future to he request.cookie
        # so that we can mature the future when the controls are set
        controls = self.controls.get_libcamera_controls()
        for id, value in controls.items():
            request.set_control(id, value)
        self.controls = Controls(self)

        code = self.camera.queue_request(request)
        errno_handle(code, f"camera.queue_request({request})")

    def _make_requests(self) -> List[libcamera.Request]:
        """Make libcamera request objects.

        Makes as many as the number of buffers in the stream with the smallest number of buffers.

        :raises RuntimeError: Failure
        :return: requests
        :rtype: List[libcamera.Request]
        """
        num_requests = min(
            [len(self.allocator.buffers(stream)) for stream in self.streams]
        )
        requests = []
        for i in range(num_requests):
            request = self.camera.create_request(self.camera_idx)
            if request is None:
                raise RuntimeError("Could not create request")

            for stream in self.streams:
                code = request.add_buffer(stream, self.allocator.buffers(stream)[i])
                errno_handle(code, "Request.add_buffer()")
            requests.append(request)
        _log.warning("Made %d requests", len(requests))
        return requests

    def _config_opts(self, config: dict | CameraConfig) -> CameraConfig:
        if isinstance(config, CameraConfig):
            return config

        if isinstance(config, dict):
            _log.warning("Using old-style camera config, please update")
            config = config.copy()
            return CameraConfig(camera=self, **config)

        raise RuntimeError(
            f"Don't know how to make a config from {config} ({type(config)})"
        )

    def configure(self, config: dict | CameraConfig) -> None:
        """Configure the camera system with the given configuration.

        :param camera_config: Configuration, defaults to the 'preview' configuration
        :type camera_config: dict, string or CameraConfiguration, optional
        :raises RuntimeError: Failed to configure
        """
        if self.started:
            raise RuntimeError("Camera must be stopped before configuring")
        camera_config = self._config_opts(config)

        # Mark ourselves as unconfigured.
        self.camera_config = None

        # Check the config and turn it into a libcamera config.
        camera_config.apply(self.camera)
        self.stream_map = camera_config.get_stream_map()

        # Update the controls and properties list as some of the values may have changed.
        self.camera_ctrl_info = lc_unpack_controls(self.camera.controls)
        self.camera_properties_ = lc_unpack(self.camera.properties)

        # Allocate all the frame buffers.
        self.streams = [
            stream_config.stream for stream_config in camera_config.libcamera_config
        ]

        # TODO(meawoppl) - can be taken off public and used in the 1 function
        # that calls it.
        self.allocator = libcamera.FrameBufferAllocator(self.camera)
        for i, stream in enumerate(self.streams):
            if self.allocator.allocate(stream) < 0:
                _log.critical("Failed to allocate buffers.")
                raise RuntimeError("Failed to allocate buffers.")
            msg = f"Allocated {len(self.allocator.buffers(stream))} buffers for stream {i}."
            _log.debug(msg)
        # Mark ourselves as configured.
        self.camera_config = camera_config

        # Set the controls directly so as to overwrite whatever is there.
        self.controls.set_controls(self.camera_config.controls)

    def camera_configuration(self) -> CameraConfig:
        """Return the camera configuration."""
        return self.camera_config

    def stream_configuration(self, name="main") -> dict:
        """Return the stream configuration for the named stream."""
        return self.camera_config[name]

    def _start(self) -> None:
        """Start the camera system running."""
        if self.camera_config is None:
            raise RuntimeError("Camera has not been configured")
        if self.started:
            raise RuntimeError("Camera already started")
        controls = self.controls.get_libcamera_controls()
        self.controls = Controls(self)

        code = self.camera.start(controls)
        errno_handle(code, "camera.start()")

        for request in self._make_requests():
            self.camera.queue_request(request)
        self.started = True
        _log.info("Camera started")

    def start(self) -> None:
        """
        Start the camera system running.
        """
        if self.camera_config is None:
            _log.warning("Camera has not been configured, using preview config")
            self.configure(CameraConfig.for_preview(self))
        if self.camera_config is None:
            raise RuntimeError("Camera has not been configured")
        # By default we will create an event loop is there isn't one running already.
        if not self._preview:
            self.start_preview()
        self._start()

    def _stop(self) -> None:
        """Stop the camera.

        Only call this function directly from within the camera event
        loop, such as in a Qt application.
        """
        if self.started:
            self.stop_count += 1
            code = self.camera.stop()
            errno_handle(code, "camera.stop()")

            # Flush Requests from the event queue.
            # This is needed to prevent old completed Requests from showing
            # up when the camera is started the next time.
            n_flushed = self._cm.handle_request(self.camera_idx)
            self.started = False
            _log.warning("Flushed %s requests", n_flushed)
            self._requests = deque()
            _log.info("Camera stopped")

    def stop(self) -> None:
        """Stop the camera."""
        if not self.started:
            _log.debug("Camera was not started")
            return
        self.stop_preview()
        self._stop()

    def set_controls(self, controls) -> None:
        """Set camera controls. These will be delivered with the next request that gets submitted."""
        self.controls.set_controls(controls)



"""Null preview"""
import threading
from logging import getLogger

_log = getLogger(__name__)


class NullPreview:
    """Null Preview"""

    def thread_func(self):
        """Thread function

        :param camera: Camera object
        :type camera: Camera
        """
        while not self._abort.is_set():
            if not self.camera.has_requests():
                self._abort.wait(0.01)
                continue

            try:
                self.camera.process_requests()
            except Exception as e:
                _log.exception("Exception during process_requests()", exc_info=e)
                raise

    def __init__(self, x=None, y=None, width=None, height=None, transform=None):
        """Initialise null preview

        :param x: X position, defaults to None
        :type x: int, optional
        :param y: Y position, defaults to None
        :type y: int, optional
        :param width: Width, defaults to None
        :type width: int, optional
        :param height: Height, defaults to None
        :type height: int, optional
        :param transform: Transform, defaults to None
        :type transform: libcamera.Transform, optional
        """
        # Ignore width and height as they are meaningless. We only accept them so as to
        # be a drop-in replacement for the Qt/DRM previews.
        self.size = (width, height)
        self._abort = threading.Event()
        self.camera = None

    def start(self, camera):
        """Starts null preview

        :param camera: Camera object
        :type camera: Camera
        """
        self.camera = camera
        self._abort.clear()
        self.thread = threading.Thread(target=self.thread_func, daemon=True)
        self.thread.start()

    def stop(self):
        """Stop preview"""
        self._abort.set()
        self.thread.join()
        self.camera = None



from __future__ import annotations

from dataclasses import asdict, dataclass, replace
from logging import getLogger
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

import libcamera

from scicamera import formats
from scicamera.controls import Controls
from scicamera.lc_helpers import errno_handle

if TYPE_CHECKING:
    from scicamera.camera import Camera

_log = getLogger(__name__)


def _assert_type(thing: Any, type_) -> None:
    if not isinstance(thing, type_):
        raise TypeError(f"{thing} should be a {type_} not {type(thing)}")


_raw_stream_ignore_list = [
    "bit_depth",
    "crop_limits",
    "exposure_limits",
    "fps",
    "unpacked",
]

STILL = libcamera.StreamRole.StillCapture
RAW = libcamera.StreamRole.Raw
VIDEO = libcamera.StreamRole.VideoRecording
VIEWFINDER = libcamera.StreamRole.Viewfinder


@dataclass
class StreamConfig:
    size: tuple[int, int]
    format: Optional[str] = None
    stride: Optional[int] = None
    framesize: Optional[int] = None

    @classmethod
    def from_lc_stream_config(cls, libcamera_stream_config):
        return cls(
            format=str(libcamera_stream_config.pixel_format),
            size=(
                libcamera_stream_config.size.width,
                libcamera_stream_config.size.height,
            ),
            stride=libcamera_stream_config.stride,
            framesize=libcamera_stream_config.frame_size,
        )

    def make_dict(self):
        return asdict(self)

    def align(self, optimal=True):
        if optimal:
            # Adjust the image size so that all planes are a mutliple of 32 bytes wide.
            # This matches the hardware behaviour and means we can be more efficient.
            align = 32
            if self.format in ("YUV420", "YVU420"):
                align = 64  # because the UV planes will have half this alignment
            elif self.format in ("XBGR8888", "XRGB8888"):
                align = (
                    16  # 4 channels per pixel gives us an automatic extra factor of 2
                )
        else:
            align = 2
        self.size = (
            self.size[0] - self.size[0] % align,
            self.size[1] - self.size[1] % 2,
        )

    def __post_init__(self) -> None:
        self.check("internal")

    def check(self, name: str):
        """Check the configuration of the passed in config.

        Raises RuntimeError if the configuration is invalid.
        """
        # Check the parameters for a single stream.
        if self.format not in [None, "MJPEG"]:
            _assert_type(self.format, str)

            if name == "raw":
                if not formats.is_raw(self.format):
                    raise RuntimeError("Unrecognized raw format " + self.format)
            else:
                if not formats.is_format_valid(self.format):
                    raise RuntimeError(
                        "Bad format " + self.format + " in stream " + name
                    )

        _assert_type(self.size, tuple)
        if len(self.size) != 2:
            raise RuntimeError(
                f"size in {name} stream should be (width, height) got: {self.size}"
            )

        for i in range(2):
            if self.size[i] % 2:
                raise RuntimeError(
                    f"All dimensions in {name} stream should be even got: {self.size}"
                )


@dataclass
class CameraConfig:
    camera: Camera
    use_case: str
    buffer_count: int
    transform: libcamera._libcamera.Transform
    color_space: libcamera._libcamera.ColorSpace

    # The are allowed to be a dict when user input, but will
    # be transformed to the proper class by the __post_init__ method.
    controls: Controls | dict
    main: StreamConfig | dict
    lores: Optional[StreamConfig | dict] = None
    raw: Optional[StreamConfig | dict] = None

    # This is set only after validation by libcamera
    libcamera_config: Optional[Any] = None

    def get_stream_config(self, name: str) -> StreamConfig | None:
        return {
            "main": self.main,
            "lores": self.lores,
            "raw": self.raw,
        }[name]

    def enable_lores(self, enable: bool = True) -> None:
        self.lores = (
            StreamConfig(size=self.main.size, format="YUV420") if enable else None
        )

    def enable_raw(self, enable: bool = True) -> None:
        self.raw = (
            StreamConfig(size=self.main.size, format=self.camera.sensor_format)
            if enable
            else None
        )

    def align(self, optimal=True):
        self.main.align(optimal)
        if self.lores is not None:
            self.lores.align(optimal)
        # No sense trying to align the raw stream.

    def get_stream_indices(self) -> Tuple[int, int, int]:
        """Get the main, lores, and raw stream indices.

        These indices will be -1 if unset.
        """
        # Get the indices of the streams we want to use.
        index = 1
        main_index = 0
        lores_index = -1
        raw_index = -1
        if self.lores is not None:
            lores_index = index
            index += 1
        if self.raw is not None:
            raw_index = index
        return main_index, lores_index, raw_index

    def __post_init__(self) -> None:
        if isinstance(self.controls, dict):
            _log.warning("CameraConfiguration controls should be a Controls object")
            self.controls = Controls(self.camera, self.controls)
        if isinstance(self.main, dict):
            _log.warning("CameraConfiguration 'main' should be a StreamConfiguration")
            self.main = StreamConfig(**self.main)
        if isinstance(self.lores, dict):
            _log.warning("CameraConfiguration 'lores' should be a StreamConfiguration")
            self.lores = StreamConfig(**self.lores)
        if isinstance(self.raw, dict):
            _log.warning("CameraConfiguration 'raw' should be a StreamConfiguration")
            self.raw = StreamConfig(**self.raw)

        # Check the entire camera configuration for errors.
        _assert_type(self.color_space, libcamera._libcamera.ColorSpace)
        _assert_type(self.transform, libcamera._libcamera.Transform)

        self.main.check("main")
        if self.lores is not None:
            self.lores.check("lores")
            main_w, main_h = self.main.size
            lores_w, lores_h = self.lores.size
            if lores_w > main_w or lores_h > main_h:
                raise RuntimeError("lores stream dimensions may not exceed main stream")
            if not formats.is_YUV(self.lores.format):
                raise RuntimeError("lores stream must be YUV")

        if self.raw is not None:
            self.raw.check("raw")

    # TODO(meawoppl) - These can likely be made static/hoisted
    @classmethod
    def for_preview(
        cls,
        camera: Camera,
        main: dict = {},
        lores=None,
        raw=None,
        transform=libcamera.Transform(),
        color_space=libcamera.ColorSpace.Sycc(),
        buffer_count=4,
        controls={},
    ) -> CameraConfig:
        """Make a configuration suitable for camera preview."""
        main_stream = StreamConfig(format="XBGR8888", size=(640, 480))
        main_stream = replace(main_stream, **main)
        main_stream.align(optimal=False)

        if lores is not None:
            lores_stream = StreamConfig(format="YUV420", size=main_stream.size)
            lores_stream = replace(lores_stream, **lores)
            lores_stream.align(optimal=False)
        else:
            lores_stream = None

        if raw is not None:
            raw_stream = StreamConfig(
                format=camera.sensor_format, size=camera.sensor_resolution
            )
            updates: dict = raw.copy()
            for name in _raw_stream_ignore_list:
                updates.pop(name, None)
            raw_stream = replace(raw_stream, **updates)
        else:
            raw_stream = None
        # Let the framerate vary from 12fps to as fast as possible.
        if (
            "NoiseReductionMode" in camera.camera_controls
            and "FrameDurationLimits" in camera.camera_controls
        ):
            controls = {
                "NoiseReductionMode": libcamera.controls.draft.NoiseReductionModeEnum.Minimal,
                "FrameDurationLimits": (100, 83333),
            } | controls
        return cls(
            camera=camera,
            use_case="preview",
            transform=transform,
            color_space=color_space,
            buffer_count=buffer_count,
            controls=controls,
            main=main_stream,
            lores=lores_stream,
            raw=raw_stream,
        )

    @classmethod
    def for_still(
        cls,
        camera,
        main={},
        lores=None,
        raw=None,
        transform=libcamera.Transform(),
        color_space=libcamera.ColorSpace.Sycc(),
        buffer_count=1,
        controls={},
    ) -> CameraConfig:
        """Make a configuration suitable for still image capture. Default to 2 buffers, as the Gl preview would need them."""
        main_stream = StreamConfig(format="BGR888", size=camera.sensor_resolution)
        main_stream = replace(main_stream, **main)
        main_stream.align(optimal=False)

        if lores is not None:
            lores_stream = StreamConfig(format="YUV420", size=main_stream.size)
            lores_stream = replace(lores_stream, **lores)
            lores_stream.align(optimal=False)
        else:
            lores_stream = None

        if raw is not None:
            raw_stream = StreamConfig(
                format=camera.sensor_format, size=main_stream.size
            )
            raw_stream = replace(raw_stream, **raw)
        else:
            raw_stream = None
        # Let the framerate span the entire possible range of the sensor.
        if (
            "NoiseReductionMode" in camera.camera_controls
            and "FrameDurationLimits" in camera.camera_controls
        ):
            controls = {
                "NoiseReductionMode": libcamera.controls.draft.NoiseReductionModeEnum.HighQuality,
                "FrameDurationLimits": (100, 1000000 * 1000),
            } | controls
        return cls(
            camera=camera,
            use_case="still",
            transform=transform,
            color_space=color_space,
            buffer_count=buffer_count,
            controls=controls,
            main=main_stream,
            lores=lores_stream,
            raw=raw_stream,
        )

    @classmethod
    def for_video(
        cls,
        camera: Camera,
        main={},
        lores=None,
        raw=None,
        transform=libcamera.Transform(),
        color_space=None,
        buffer_count=6,
        controls={},
    ) -> CameraConfig:
        """Make a configuration suitable for video recording."""
        main_stream = StreamConfig(format="XBGR8888", size=(1280, 720))
        main_stream = replace(main_stream, **main)
        main_stream.align(optimal=False)

        if lores is not None:
            lores_stream = StreamConfig(format="YUV420", size=main_stream.size)
            lores_stream = replace(lores_stream, **lores)
            lores_stream.align(optimal=False)
        else:
            lores_stream = None

        if raw is not None:
            raw_stream = StreamConfig(
                format=camera.sensor_format, size=main_stream.size
            )
            raw_stream = replace(raw_stream, **raw)
        else:
            raw_stream = None

        if color_space is None:
            # Choose default color space according to the video resolution.
            if formats.is_RGB(main_stream.format):
                # There's a bug down in some driver where it won't accept anything other than
                # sRGB or JPEG as the color space for an RGB stream. So until that is fixed:
                color_space = libcamera.ColorSpace.Sycc()
            elif main_stream.size[0] < 1280 or main_stream.size[1] < 720:
                color_space = libcamera.ColorSpace.Smpte170m()
            else:
                color_space = libcamera.ColorSpace.Rec709()
        if (
            "NoiseReductionMode" in camera.camera_controls
            and "FrameDurationLimits" in camera.camera_controls
        ):
            controls = {
                "NoiseReductionMode": libcamera.controls.draft.NoiseReductionModeEnum.Fast,
                "FrameDurationLimits": (33333, 33333),
            } | controls
        return cls(
            camera=camera,
            use_case="video",
            transform=transform,
            color_space=color_space,
            buffer_count=buffer_count,
            controls=controls,
            main=main_stream,
            lores=lores_stream,
            raw=raw_stream,
        )

    def _update_libcamera_stream_config(
        self, libcamera_stream_config, stream_config: StreamConfig, buffer_count: int
    ) -> None:
        # Update the libcamera stream config with ours.
        libcamera_stream_config.size = libcamera.Size(*stream_config.size)
        libcamera_stream_config.pixel_format = libcamera.PixelFormat(
            stream_config.format
        )
        libcamera_stream_config.buffer_count = buffer_count

    def apply(self, lc_camera) -> Dict[str, Any]:
        # Make a libcamera configuration object from our Python configuration.

        # We will create each stream with the "viewfinder" role just to get the stream
        # configuration objects, and note the positions our named streams will have in
        # libcamera's stream list.
        roles = [VIEWFINDER]
        main_index, lores_index, raw_index = self.get_stream_indices()
        if self.lores is not None:
            roles += [VIEWFINDER]
        if self.raw is not None:
            roles += [RAW]

        # Make the libcamera configuration, and then we'll write all our parameters over
        # the ones it gave us.
        libcamera_config = lc_camera.generate_configuration(roles)
        if libcamera_config is None:
            raise RuntimeError(
                f"Failed to generate libcamera configuration for {self} using {lc_camera}"
            )
        libcamera_config.transform = self.transform
        self._update_libcamera_stream_config(
            libcamera_config.at(main_index), self.main, self.buffer_count
        )
        libcamera_config.at(main_index).color_space = self.color_space
        if self.lores is not None:
            self._update_libcamera_stream_config(
                libcamera_config.at(lores_index),
                self.lores,
                self.buffer_count,
            )
            libcamera_config.at(lores_index).color_space = self.color_space

        if self.raw is not None:
            self._update_libcamera_stream_config(
                libcamera_config.at(raw_index), self.raw, self.buffer_count
            )
            libcamera_config.at(raw_index).color_space = libcamera.ColorSpace.Raw()

        # Check that libcamera is happy with it.
        _log.debug(f"Validating configuration: {self}")
        status = libcamera_config.validate()
        if status == libcamera.CameraConfiguration.Status.Invalid:
            raise RuntimeError(f"Invalid camera configuration: {self}")
        elif status == libcamera.CameraConfiguration.Status.Adjusted:
            _log.info("Camera configuration has been adjusted!")

        # Back prop any tweaks it made onto this class
        self.transform = libcamera_config.transform
        self.color_space = libcamera_config.at(0).color_space
        self.main = StreamConfig.from_lc_stream_config(libcamera_config.at(0))
        if lores_index >= 0:
            self.lores = StreamConfig.from_lc_stream_config(
                libcamera_config.at(lores_index)
            )
        if raw_index >= 0:
            self.raw = StreamConfig.from_lc_stream_config(
                libcamera_config.at(raw_index)
            )

        # Configure the lc_camera instance
        code = lc_camera.configure(libcamera_config)
        errno_handle(code, "camera.configure()")
        _log.info("Configuration successful!")
        _log.debug(f"Final configuration: {self}")

        self.libcamera_config = libcamera_config

    def get_stream_map(self) -> Dict[str, Any]:
        # Record which libcamera stream goes with which of our names.
        stream_map = {}
        for idx, name in zip(self.get_stream_indices(), ("main", "lores", "raw")):
            if idx >= 0:
                stream_map[name] = self.libcamera_config.at(idx).stream
        return stream_map



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scicamera']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.23.5,<2.0.0', 'pillow>=9.3.0,<10.0.0']

setup_kwargs = {
    'name': 'scicamera',
    'version': '0.4.0',
    'description': '',
    'long_description': '# SciCamera\n\n---\nConsistent and reliable imaging for scientific applications.\n\n\n## Why _SciCamera_?\n\nScientific imaging applications often require minimal post-processing pipelines, \nprecise capture timing, near-gapless sequential frames, and easily \nconfigurable settings like gain, resolution, bit-depth, and exposure \nlength. \n\nThis project, which began as fork of the webcam/video-focused [`picamera2`][picamera2]\nlibrary, aims to make it easy to configure and use cameras for scientific applications,\nwith a focus on _performance, reliability, code quality, and maintainability_.\n\n\n### Why not _SciCamera_?\n\nSciCamera currently focuses on high-quality, timing-sensitive, minimally-processed\n_still images_. For low-bandwidth, real-time image and video streaming, we recommend \nthe [`picamera2`][picamera2] library.\n\n\n## Platform support\n\n_SciCamera_ supports\n\n- Raspberry Pi OS (Bullseye or later), 64-bit.\n- x86 Ubuntu\n\nOther debian flavors are likely to be supported. We welcome pull requests to extend\nthe testing toolchains to cover your platform.\n\n## Installation\n\n_SciCamera_ is a pure python package, but relies on the python\nc++ wrapper of _libcamera_.\n\n_SciCamera_ can be installed simply with:\n```\npip install scicamera\n```\n\n### Installing libcamera + python bindings\n\nImport and use of the above pacakge requires that `libcamera` to be built\nwith the python package enabled. On rasbian, this is accomplished by \ninstalling the `libcamera` package from apt. In x86 it must be built \nusing something like the following:\n\n```bash\ngit clone https://github.com/Exclosure/libcamera.git\ncd libcamera\ngit checkout v0.0.4\nmeson setup build -D pycamera=enabled\nninja -C build\nsudo ninja -C build install\n```\n\n## Bugs/Contributing\n\n\nOpen an issue/PR to discuss your bug or feature. Once a course of action\nhas been identified, open a PR, discuss the changes. \n\nFeature creep is not of interest, but we would be happy\nto help you build your more complicated project on top of this.\n\nIf we like them, and the tests pass we will merge them. \nCI requires code has been processed `isort` and `black` toolchains.\n\nDoing this is pretty easy:\n```\nisort .\nblack .\n```\n\nGreat work.\n\n## Publishing to PYPI\n\nShould be added to github action later\n\n1. Add your pypi token\n  ```sh\n  $ poetry config pypi-token.pypi my-token\n  ```\n\n2. Cut a new tag\n  ```sh\n  $ git tag -a v0.1.0 -m "Version 0.1.0"\n  $ git push origin v0.1.0\n  ```\n\n3. Publish\n  ```sh\n  $ poetry publish --build\n  ```\n\n\n[picamera2]:https://github.com/raspberrypi/picamera2\n',
    'author': 'Exclosure',
    'author_email': 'info@exclosure.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/exclosure/scicamera',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import json
import os
import tempfile


def load_tuning_file(tuning_file: str, dir=None):
    """Load the named tuning file.

    If dir is given, then only that directory is checked,
    otherwise a list of likely installation directories is searched

    :param tuning_file: Tuning file
    :type tuning_file: str
    :param dir: Directory of tuning file, defaults to None
    :type dir: str, optional
    :raises RuntimeError: Produced if tuning file not found
    :return: Dictionary of tuning file
    :rtype: dict
    """
    if dir is not None:
        dirs = [dir]
    else:
        dirs = [
            "/home/pi/libcamera/src/ipa/raspberrypi/data",
            "/usr/local/share/libcamera/ipa/raspberrypi",
            "/usr/share/libcamera/ipa/raspberrypi",
        ]
    for dir in dirs:
        file = os.path.join(dir, tuning_file)
        if os.path.isfile(file):
            with open(file, "r") as fp:
                return json.load(fp)
    raise RuntimeError("Tuning file not found")


def find_tuning_algo(tuning: dict, name: str) -> dict:
    """
    Return the parameters for the named algorithm in the given camera tuning.

    :param tuning: The camera tuning object
    :type tuning: dict
    :param name: The name of the algorithm
    :type name: str
    :rtype: dict
    """
    version = tuning.get("version", 1)
    if version == 1:
        return tuning[name]
    return next(algo for algo in tuning["algorithms"] if name in algo)[name]


class TuningContext:
    def __init__(self, tuning: str | dict | None):
        self.tuning = tuning

    def __enter__(self):
        self._tuning_file = None
        if self.tuning is None:
            os.environ.pop("LIBCAMERA_RPI_TUNING_FILE", None)  # Use default tuning
            return

        if isinstance(self.tuning, str):
            os.environ["LIBCAMERA_RPI_TUNING_FILE"] = self.tuning
        else:
            self._tuning_file = tempfile.NamedTemporaryFile("w")
            json.dump(self.tuning, self._tuning_file)
            self._tuning_file.flush()  # but leave it open as closing it will delete it
            os.environ["LIBCAMERA_RPI_TUNING_FILE"] = self._tuning_file.name

    def __exit__(self, *args):
        if self._tuning_file is not None:
            self._tuning_file.close()  # delete the temporary file



import errno
from logging import getLogger
from typing import Any, Dict

import libcamera

_log = getLogger(__name__)


def errno_handle(code: int, callname: str) -> None:
    """Convert an errno to a string.

    If code is >= 0, it will log a debug message and otherwise noop
    If code is < 0, it will log an error message and raise a ``RuntimeError``
    """
    if code >= 0:
        str_rep = "Success"
    else:
        str_rep = errno.errorcode.get(-code, "Unknown error")
    formatted = f"{callname} - {str_rep} ({code})"

    if code >= 0:
        _log.debug(formatted)
    else:
        _log.error(formatted)
        raise RuntimeError(formatted)


def _convert_from_libcamera_type(value):
    if isinstance(value, libcamera.Rectangle):
        value = (value.x, value.y, value.width, value.height)
    elif isinstance(value, libcamera.Size):
        value = (value.width, value.height)
    elif isinstance(value, (tuple, list)):
        value = [_convert_from_libcamera_type(v) for v in value]
    return value


def lc_unpack(lc_dict) -> Dict[str, Any]:
    unpacked = {}
    for k, v in lc_dict.items():
        unpacked[k.name] = _convert_from_libcamera_type(v)
    return unpacked


def lc_unpack_controls(lc_dict) -> Dict[str, Any]:
    unpacked = {}
    for k, v in lc_dict.items():
        unpacked[k.name] = (k, v)
    return unpacked


LCTransform = libcamera._libcamera.Transform


def libcamera_transforms_eq(t1: LCTransform, t2: LCTransform) -> bool:
    """Return ``True`` if the two transforms are equivalent."""
    return (
        t1.hflip == t2.hflip and t1.vflip == t2.vflip and t1.transpose == t2.transpose
    )


def libcamera_color_spaces_eq(c1, c2) -> bool:
    """Return ``True`` if the two color spaces are equivalent."""
    return (
        c1.primaries == c2.primaries
        and c1.transferFunction == c2.transferFunction
        and c1.ycbcrEncoding == c2.ycbcrEncoding
        and c1.range == c2.range
    )
